"""Define the core `Entity` class.

Defines the core `Entity` class to link physical quantities to ontology concepts. Also
includes helper functions for inferring the correct SI units from the ontology.

"""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

import mammos_units as u

from mammos_entity._onto import mammos_ontology

if TYPE_CHECKING:
    import astropy.units
    import mammos_units
    import numpy.typing
    import owlready2

    import mammos_entity


base_units = [u.T, u.J, u.m, u.A, u.radian, u.kg, u.s, u.K, u.mol, u.cd, u.V]


def si_unit_from_list(list_cls: list[owlready2.entity.ThingClass]) -> str:
    """Return an SI unit from a list of entities from the EMMO ontology.

    Given a list of ontology classes, determine which class corresponds to
    a coherent SI derived unit (or if none found, an SI dimensional unit),
    then return that class's UCUM code.

    Args:
        list_cls: A list of ontology classes.

    Returns:
        The UCUM code (e.g., "J/m^3", "A/m") for the first identified SI unit
        in the given list of classes.

    """
    si_unit_cls = [
        cls
        for cls in list_cls
        if mammos_ontology.SICoherentDerivedUnit in cls.ancestors()
    ]
    if not si_unit_cls:
        si_unit_cls = [
            cls
            for cls in list_cls
            if (mammos_ontology.SIDimensionalUnit in cls.ancestors())
        ]

    # Explanation of the following line:
    # 1. We find all ucum (Unified Code for Units of Measure) Code for all units
    #    in si_unit_cls.
    # 2. Astropy complains if it sees unit strings with parentheses, so we exclude
    #    them.
    # 3. We take the first item. It is not important what unit we are selecting
    #    because the ontology does not define a single preferred unit. We are
    #    taking one of the SI coherent derived units or a SI dimensional unit.
    #    astropy will make the conversion to base units later on.
    return [
        unit
        for si_unit_cls_i in si_unit_cls
        for unit in si_unit_cls_i.ucumCode
        if "(" not in unit
    ][0]


def extract_SI_units(ontology_label: str) -> str | None:
    """Find SI unit for the given label from the EMMO ontology.

    Given a label for an ontology concept, retrieve the corresponding SI unit
    by traversing the class hierarchy. If a valid unit is found, its UCUM code
    is returned; otherwise, an empty string is returned (equivalent to dimensionless).

    Args:
        ontology_label: The label of an ontology concept
            (e.g., 'SpontaneousMagnetization').

    Returns:
        The UCUM code of the concept's SI unit, or None if no suitable SI unit
        is found or if the unit is a special case like 'Cel.K-1'.

    """
    thing = mammos_ontology.get_by_label(ontology_label)
    si_unit = ""
    for ancestor in thing.ancestors():
        if hasattr(ancestor, "hasMeasurementUnit") and ancestor.hasMeasurementUnit:
            if sub_class := list(ancestor.hasMeasurementUnit[0].subclasses()):
                si_unit = si_unit_from_list(sub_class)
            elif ontology_label := ancestor.hasMeasurementUnit[0].ucumCode:
                si_unit = ontology_label[0]
            break
    # HACK: filter Celsius values as Kelvin and `Cel.K-1` as no units
    if si_unit in {"Cel", "mCel"}:
        si_unit = "K"
    elif si_unit == "Cel.K-1":
        si_unit = ""
    return si_unit


class Entity:
    """Create a quantity (a value and a unit) linked to the EMMO ontology.

    Represents a physical property or quantity that is linked to an ontology
    concept. It enforces unit compatibility with the ontology.

    Args:
        ontology_label: Ontology label
        value: Value
        unit: Unit

    Examples:
        >>> import mammos_entity as me
        >>> import mammos_units as u
        >>> Ms = me.Entity(ontology_label='SpontaneousMagnetization', value=8e5, unit='A / m')
        >>> H = me.Entity("ExternalMagneticField", 1e4 * u.A / u.m)
        >>> Tc_kK = me.Entity("CurieTemperature", 0.1, unit=u.kK)
        >>> Tc_K = me.Entity("CurieTemperature", Tc_kK, unit=u.K)

    """  # noqa: E501

    def __init__(
        self,
        ontology_label: str,
        value: numpy.typing.ArrayLike
        | mammos_units.Quantity
        | mammos_entity.Entity = 0,
        unit: str | None | mammos_units.UnitBase = None,
    ):
        if isinstance(value, Entity):
            if value.ontology_label != ontology_label:
                raise ValueError(
                    "Incompatible label for initialization."
                    f" Trying to initialize a {ontology_label}"
                    f" with a {value.ontology_label}."
                )
            value = value.quantity

        if unit is None and isinstance(value, u.Quantity):
            unit = value.unit

        si_unit = extract_SI_units(ontology_label)

        if (si_unit is not None) and (unit is not None):
            # Remove any set equivalency to enforce unit strictness
            with u.set_enabled_equivalencies(None):
                if not u.Unit(si_unit).is_equivalent(unit):
                    raise u.UnitConversionError(
                        f"The unit '{unit}' is not equivalent to the unit of"
                        f" {ontology_label} '{u.Unit(si_unit)}'"
                    )
        elif (si_unit is not None) and (unit is None):
            with u.add_enabled_aliases({"Cel": u.K, "mCel": u.K}):
                comp_si_unit = u.Unit(si_unit).decompose(bases=base_units)
            unit = u.CompositeUnit(1, comp_si_unit.bases, comp_si_unit.powers)
        elif (si_unit is None) and unit:
            raise TypeError(
                f"{ontology_label} is a unitless entity."
                f" Hence, {unit} is inappropriate."
            )

        comp_unit = u.Unit(unit if unit else "")

        # Remove any set equivalency to enforce unit strictness
        with u.set_enabled_equivalencies(None):
            self._quantity = u.Quantity(value=value, unit=comp_unit)
        self._ontology_label = ontology_label

    @property
    def ontology_label(self) -> str:
        """The ontology label that links the entity to the EMMO ontology.

        Retrieve the ontology label corresponding to the `ThingClass` that defines the
        given entity in ontology.

        Returns:
            str: The ontology label corresponding to the right ThingClass.

        """
        return self._ontology_label

    @property
    def ontology_label_with_iri(self) -> str:
        """The ontology label with its IRI. Unique link to EMMO ontology.

        Returns the `self.ontology_label` together with the IRI (a URL that
        points to the definition of this entity.) IRI stands for
        Internationalized Resource Identifier.

        If only the IRI is desired, one can use `self.ontology.iri`.

        Returns:
            str: The ontology label corresponding to the right ThingClass,
                 together with the IRI.

        """
        return f"{self.ontology_label} {self.ontology.iri}"

    # FIX: right not this will fail if no internet!
    @property
    def ontology(self) -> owlready2.entity.ThingClass:
        """Retrieve the ontology class corresponding to the entity's label.

        Returns:
            The ontology class from `mammos_ontology` that matches the entity's label.

        """
        return mammos_ontology.get_by_label(self.ontology_label)

    @property
    def quantity(self) -> astropy.units.Quantity:
        """Return the entity as a `mammos_units.Quantity`.

        Return a stand-alone `mammos_units.Quantity` object with the same value
        and unit, detached from the ontology link.

        Returns:
            A copy of this entity as a pure physical quantity.

        """
        return self._quantity

    @property
    def q(self) -> mammos_units.Quantity:
        """Quantity attribute, shorthand for `.quantity`."""
        return self.quantity

    @property
    def value(self) -> numpy.scalar | numpy.ndarray:
        """Numerical data of the entity."""
        return self.quantity.value

    @property
    def unit(self) -> astropy.units.UnitBase:
        """Unit of the entity data."""
        return self.quantity.unit

    @property
    def axis_label(self) -> str:
        """Return an ontology-based axis label for the plots.

        The axis label consist of ontology label and unit:
        - The ontology label is split with spaces at all capital letters
        - The units are added in parentheses.

        Returns:
            A string for labelling the axis corresponding to the entity on a plot.

        Examples:
            >>> import mammos_entity as me
            >>> me.Entity("SpontaneousMagnetization").axis_label
            'Spontaneous Magnetization (A / m)'
            >>> me.Entity("DemagnetizingFactor").axis_label
            'Demagnetizing Factor'
        """
        return re.sub(r"(?<!^)(?=[A-Z])", " ", f"{self.ontology_label}") + (
            f" ({self.unit})" if str(self.unit) else ""
        )

    def __eq__(self, other: mammos_entity.Entity) -> bool:
        """Check if two Entities are identical.

        Entities are considered identical if they have the same ontology label and
        numerical data, i.e. unit prefixes have no effect.

        Examples:
            >>> import mammos_entity as me
            >>> ms_1 = me.Ms(1, "kA/m")
            >>> ms_2 = me.Ms(1e3, "A/m")
            >>> ms_1 == ms_2
            True
            >>> t = me.T(1, "K")
            >>> ms_1 == t
            False
        """
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.ontology_label == other.ontology_label
            and self.q.shape == other.q.shape
            and u.allclose(self.q, other.q)
        )

    def __repr__(self) -> str:
        args = [f"ontology_label='{self._ontology_label}'", f"value={self.value!r}"]
        if str(self.unit):
            args.append(f"unit='{self.unit!s}'")

        return f"{self.__class__.__name__}({', '.join(args)})"

    def __str__(self) -> str:
        new_line = "\n" if self.value.size > 4 else ""
        if self.unit.is_equivalent(u.dimensionless_unscaled):
            repr_str = f"{self.ontology_label}(value={new_line}{self.value})"
        else:
            repr_str = (
                f"{self.ontology_label}(value={new_line}{self.value}"
                f",{new_line} unit={self.unit})"
            )
        return repr_str

    def _repr_html_(self) -> str:
        html_str = str(self).replace("\n", "<br>").replace(" ", "&nbsp;")
        return f"<samp>{html_str}</samp>"



import math

import mammos_units as u
import numpy as np
import pytest
from numpy import array  # noqa: F401  # required for repr eval

import mammos_entity as me
from mammos_entity import Entity  # noqa: F401  # required for repr eval


def test_init_float():
    """Initialize Entity instance with a float."""
    e = me.Entity("ExternalMagneticField", value=8e5)
    q = 8e5 * u.A / u.m
    assert u.allclose(e.quantity, q)
    assert np.allclose(e.value, 8e5)
    assert e.unit == u.A / u.m
    assert e.ontology_label == "ExternalMagneticField"


def test_init_list():
    """Initialize with Python lists."""
    val = [42, 42, 42]
    e = me.Entity("ExternalMagneticField", value=val)
    assert np.allclose(e.value, val)
    val[0] = 1
    assert np.allclose(e.value, [42, 42, 42])


def test_init_tuple():
    """Initialize with Python tuples."""
    val = (42, 42, 42)
    e = me.Entity("ExternalMagneticField", value=val)
    assert np.allclose(e.value, np.array(val))


def test_init_numpy():
    """Initialize with NumPy array."""
    val = np.array([42, 42, 42])
    e = me.Entity("ExternalMagneticField", value=val)
    assert np.allclose(e.value, val)
    val[0] = 1
    assert np.allclose(e.value, [42, 42, 42])
    val = np.ones((42, 42, 42, 3))
    e = me.Entity("ExternalMagneticField", value=val)
    assert np.allclose(e.value, val)


def test_init_quantity():
    """Initialize using mammos_units.Quantity.

    Test 1: an entity created from a quantity without specifying unit
    will take value and unit from the quantity. In this case the unit
    of the quantity is the default ontology quantity.
    Test 2: an entity created from a quantity specifying the unit
    will convert the quantity to the selected unit. In this case
    the unit is the same of the quantity, so there is actually no
    conversion involved.
    Test 3: Same as Test 1, but this time the unit of the quantity
    is not the default ontology quantity.
    Test 4: Same as Test 2, but there is an actually conversion involved.
    """
    q = 1 * u.A / u.m
    e = me.Entity("ExternalMagneticField", value=q)
    assert e.ontology_label == "ExternalMagneticField"
    assert u.allclose(e.quantity, q)
    assert np.allclose(e.value, 1)
    assert e.unit == u.A / u.m
    q = 1 * u.kA / u.m
    e = me.Entity("ExternalMagneticField", value=q, unit="kA/m")
    assert u.allclose(e.quantity, q)
    assert np.allclose(e.value, 1)
    assert e.unit == u.kA / u.m
    e = me.Entity("ExternalMagneticField", value=q)
    assert u.allclose(e.quantity, q)
    assert np.allclose(e.value, 1)
    assert e.unit == u.kA / u.m
    e = me.Entity("ExternalMagneticField", value=q, unit="MA/m")
    assert u.allclose(e.quantity, q)
    assert np.allclose(e.value, 1e-3)
    assert e.unit == u.MA / u.m


def test_init_entity():
    """Initialize from another Entity.

    Test 1: an Entity initialized from another Entity will define
    its Quantity (including unit) from it.
    Test 2: if we select a different unit, it gets converted.
    Test 3: if we initialize using an Entity with a different ontology label
    we get an error.
    """
    e_1 = me.Entity("ExternalMagneticField", value=1, unit="mA/m")
    e_2 = me.Entity("ExternalMagneticField", value=e_1)
    assert e_2.ontology_label == "ExternalMagneticField"
    assert u.allclose(e_1.quantity, e_2.quantity)
    assert np.allclose(e_1.value, e_2.value)
    assert e_1.unit == e_2.unit
    e_3 = me.Entity("ExternalMagneticField", value=e_1, unit="A/m")
    assert u.allclose(e_3.quantity, e_1.quantity)
    assert np.allclose(e_3.value, 1e-3)
    assert e_3.unit == u.A / u.m
    with pytest.raises(ValueError):
        me.Entity("CurieTemperature", value=e_1)


def test_unitless():
    """Test unitless Entity."""
    e_1 = me.Entity("DemagnetizingFactor", 0.3)
    assert e_1.ontology_label == "DemagnetizingFactor"
    assert math.isclose(e_1.value, 0.3)
    assert e_1.unit.is_equivalent("")
    e_2 = me.Entity("DemagnetizingFactor", [1, 2])
    assert np.allclose(e_2.value, [1, 2])
    assert e_2.unit.is_equivalent("")
    e_3 = me.Entity("DemagnetizingFactor", u.Quantity(0.3))
    assert math.isclose(e_3.value, 0.3)
    assert e_3.unit.is_equivalent("")
    e_4 = me.Entity("DemagnetizingFactor", e_3)
    assert math.isclose(e_4.value, 0.3)
    assert e_4.unit.is_equivalent("")


def test_check_units():
    """Test units of Entity.

    Test 1: Check that unit is immutable.
    Test 2: Check that Entity cannot be initialized with wrong unit.
    Even if we activate the necessary conversion equivalency, the initialization
    should reset all equivalencies.
    """
    # change unit (conversion/change unit after initialized entity)
    e = me.Entity("SpontaneousMagnetization", value=1, unit=u.A / u.m)
    e.quantity.to("kA/m")
    assert e.unit == u.A / u.m
    e.quantity.to("kA/m", copy=False)
    assert e.unit == u.A / u.m
    with pytest.raises(u.UnitConversionError):
        me.Entity("SpontaneousMagnetization", value=1, unit="T")
    with (
        u.set_enabled_equivalencies(u.magnetic_flux_field()),
        pytest.raises(u.UnitConversionError),
    ):
        me.Entity("SpontaneousMagnetization", value=1, unit="T")
    with (
        u.set_enabled_equivalencies(u.magnetic_flux_field()),
        pytest.raises(u.UnitConversionError),
    ):
        me.Entity("SpontaneousMagnetization", value=1 * u.T, unit="A/m")


def test_repr():
    """Test representation string.

    Test 1: Test repr for scalar value.
    Test 2: Test repr for vectorial value.
    Test 3: Test repr for unitless Entity.
    """
    e = me.Entity("CurieTemperature")
    assert (
        e.__repr__()
        == "Entity(ontology_label='CurieTemperature', value=np.float64(0.0), unit='K')"
    )
    assert eval(repr(e)) == e

    a = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    e = me.Entity("ExternalMagneticField", value=a)
    assert e.__repr__() == (
        "Entity(ontology_label='ExternalMagneticField', "
        + f"value={np.array(a, dtype=float)!r}, unit='A / m')"
    )
    assert eval(repr(e)) == e

    e = me.Entity("DemagnetizingFactor")
    assert (
        e.__repr__()
        == "Entity(ontology_label='DemagnetizingFactor', value=np.float64(0.0))"
    )
    assert eval(repr(e)) == e


def test_axis_labels():
    """Test different axis_label examples."""
    e_1 = me.Entity("ExternalMagneticField")
    assert e_1.axis_label == "External Magnetic Field (A / m)"
    e_2 = me.Entity("AffinityOfAChemicalReaction")
    assert e_2.axis_label == "Affinity Of A Chemical Reaction (J / mol)"
    e_3 = me.Entity("DemagnetizingFactor")
    assert e_3.axis_label == "Demagnetizing Factor"
    e_4 = me.Entity("Entropy")
    assert e_4.axis_label == "Entropy (J / K)"
    # e_5 = me.Entity("PlanckConstant")
    # assert e_5.axis_label == "Planck Constant (m2 kg / s)"


@pytest.mark.parametrize("ontology_element", me.mammos_ontology.classes(imported=True))
def test_all_labels_ontology(ontology_element):
    """Test all labels in the ontology.

    This test creates one Entity instance for each label in the ontology.
    """
    me.Entity(ontology_element.prefLabel[0], 42)


def test_ontology_label_mammos():
    """Test ontology label for an Entity in the MaMMoS ontology."""
    e = me.Entity("ExternalMagneticField")
    assert e.ontology_label == "ExternalMagneticField"
    assert (
        e.ontology_label_with_iri
        == "ExternalMagneticField https://w3id.org/emmo/domain/magnetic_material#EMMO_da08f0d3-fe19-58bc-8fb6-ecc8992d5eb3"
    )
    assert e.ontology_label_with_iri == f"{e.ontology.prefLabel[0]} {e.ontology.iri}"
    assert e.ontology_label in me.mammos_ontology
    H = me.mammos_ontology.get_by_label(e.ontology_label)
    assert e.ontology_label_with_iri == f"{H.prefLabel[0]} {H.iri}"


def test_ontology_label_EMMO():
    """Test ontology label for an Entity in the EMMO."""
    e = me.Entity("AngularVelocity")
    assert e.ontology_label == "AngularVelocity"
    assert (
        e.ontology_label_with_iri
        == "AngularVelocity https://w3id.org/emmo#EMMO_bd325ef5_4127_420c_83d3_207b3e2184fd"
    )
    assert e.ontology_label_with_iri == f"{e.ontology.prefLabel[0]} {e.ontology.iri}"
    assert e.ontology_label in me.mammos_ontology
    omega = me.mammos_ontology.get_by_label(e.ontology_label)
    assert e.ontology_label_with_iri == f"{omega.prefLabel[0]} {omega.iri}"


def test_equality():
    """Test equality.

    We expect two entities to be equal if the ontology_label is the same
    and the values are close enough.
    Equality fails when the right hand term is not an Entity.
    """
    e_1 = me.Entity("SpontaneousMagnetization", value=1)
    e_2 = me.Entity("SpontaneousMagnetization", value=1)
    assert e_1 == e_2
    e_3 = me.Entity("SpontaneousMagnetization", value=2)
    assert e_1 != e_3
    e_4 = me.Entity("ExternalMagneticField", value=1)
    assert e_1 != e_4
    e_5 = me.Entity("SpontaneousMagnetization", value=1000, unit=u.mA / u.m)
    assert e_1 == e_5
    e_6 = me.Entity("SpontaneousMagnetization", value=[[1, 1]])
    assert e_1 != e_6
    e_7 = me.Entity("SpontaneousMagnetization", value=[[1], [1]])
    assert e_6 != e_7

    # Other objects
    assert e_1 != 1 * u.A / u.m
    assert e_1 != 1
    assert e_1 != e_2.quantity

    # Other objects can implement __eq__ in a way that is compatible with Entity

    class A:
        def __eq__(self, o):
            return True

    assert e_1 == A()


@pytest.mark.parametrize(
    "function, expected_label",
    (
        (me.A, "ExchangeStiffnessConstant"),
        (me.BHmax, "MaximumEnergyProduct"),
        (me.B, "MagneticFluxDensity"),
        (me.H, "ExternalMagneticField"),
        (me.Hc, "CoercivityHcExternal"),
        (me.J, "MagneticPolarisation"),
        (me.Js, "SpontaneousMagneticPolarisation"),
        (me.Ku, "UniaxialAnisotropyConstant"),
        (me.M, "Magnetization"),
        (me.Mr, "Remanence"),
        (me.Ms, "SpontaneousMagnetization"),
        (me.T, "ThermodynamicTemperature"),
        (me.Tc, "CurieTemperature"),
    ),
)
def test_known_labels(function, expected_label):
    """Check predefined entities."""
    assert function().ontology_label == expected_label



import textwrap

import mammos_units as u
import numpy as np
import pandas as pd
import pytest

import mammos_entity as me
from mammos_entity.io import EntityCollection, entities_from_csv, entities_to_csv


def test_to_csv_no_data():
    with pytest.raises(RuntimeError):
        entities_to_csv("test.csv")


@pytest.mark.skip(reason="Allow multiple datatypes in one column for now.")
def test_different_types_column():
    with pytest.raises(TypeError):
        entities_to_csv("test.csv", data=[1, me.A()])


@pytest.mark.parametrize(
    "data",
    [
        {"A": 1.0, "Ms": 2.0, "Ku": 3.0},
        {
            "A": 1.0 * (u.J / u.m),
            "Ms": 2 * (u.A / u.m),
            "Ku": 3 * (u.J / u.m**3),
        },
        {"A": me.A(1), "Ms": me.Ms(2), "Ku": me.Ku(3)},
    ],
    ids=["floats", "quantites", "entities"],
)
def test_scalar_column(tmp_path, data):
    entities_to_csv(tmp_path / "test.csv", **data)

    read_csv = entities_from_csv(tmp_path / "test.csv")

    assert data["A"] == read_csv.A
    assert data["Ms"] == read_csv.Ms
    assert data["Ku"] == read_csv.Ku


def test_read_collection_type(tmp_path):
    entities_to_csv(tmp_path / "simple.csv", data=[1, 2, 3])
    read_csv = entities_from_csv(tmp_path / "simple.csv")
    assert isinstance(read_csv, EntityCollection)
    assert np.allclose(read_csv.data, [1, 2, 3])


def test_read_write_csv(tmp_path):
    Ms = me.Ms([1e6, 2e6, 3e6])
    T = me.T([1, 2, 3])
    theta_angle = [0, 0.5, 0.7] * u.rad
    demag_factor = me.Entity("DemagnetizingFactor", [1 / 3, 1 / 3, 1 / 3])
    comments = ["Some comment", "Some other comment", "A third comment"]
    entities_to_csv(
        tmp_path / "example.csv",
        Ms=Ms,
        T=T,
        angle=theta_angle,
        n=demag_factor,
        comment=comments,
    )

    read_csv = entities_from_csv(tmp_path / "example.csv")

    assert read_csv.Ms == Ms
    assert read_csv.T == T
    # Floating-point comparisons with == should ensure that we do not loose precision
    # when writing the data to file.
    assert all(read_csv.angle == theta_angle)
    assert read_csv.n == demag_factor
    assert all(read_csv.comment == comments)

    df_with_units = read_csv.to_dataframe()
    assert list(df_with_units.columns) == [
        "Ms (A / m)",
        "T (K)",
        "angle (rad)",
        "n",
        "comment",
    ]

    df_without_units = read_csv.to_dataframe(include_units=False)
    assert list(df_without_units.columns) == ["Ms", "T", "angle", "n", "comment"]

    df = pd.read_csv(tmp_path / "example.csv", comment="#")

    assert all(df == df_without_units)


def test_wrong_file_version(tmp_path):
    file_content = textwrap.dedent(
        """
        #mammos csv v0
        #
        #
        #
        index
        1
        2
        """
    )
    (tmp_path / "data.csv").write_text(file_content)

    with pytest.raises(RuntimeError):
        me.io.entities_from_csv(tmp_path / "data.csv")



r"""Support for reading and writing Entity files.

Currently only a single file format is supported: a CSV file with additional
commented metadata lines. Comments start with #.

- The first line is commented and contains the file version in the form::

     mammos csv v<VERSION>

  The reading code checks the version number (using regex v\d+) to ensure compatibility.
- The second line is commented and contains the preferred ontology label.
- The third line is commented and contains the ontology IRI.
- The fourth line is commented and contains units.
- The fifth line contains the short labels used to refer to individual columns when
  working with the data, e.g. in a :py:class:`pandas.DataFrame`. Omitting spaces in this
  string is advisable.

  Ideally this string is the short ontology label.
- All remaining lines contain data

Elements in a line are separated by a comma without any surrounding whitespace. A
trailing comma is not permitted.

If a column has no ontology entry lines 1 and 2 are empty for this column.

If a column has no units (with or without ontology entry) line 3 has no entry for this
column.

Here is an example with five columns, an index with no units or ontology label, the
entity spontaneous magnetization with an entry in the ontology, a made-up quantity alpha
with a unit but no ontology label, demagnetizing factor with an ontology entry but no
unit, and a column `description` containing a string description without units or
ontology label. To keep this example short the actual IRIs are omitted::

   #mammos csv v1
   #,SpontaneousMagnetization,,DemagnetizingFactor,description
   #,https://w3id.org/emm/...,,https://w3id.org/emmo/...,
   #,kA/m,s^2,,
   index,Ms,alpha,DemagnetizingFactor,
   0,1e2,1.2,1,Description of the first data row
   1,1e2,3.4,0.5,Description of the second data row
   2,1e2,5.6,0.5,Description of the third data row

"""

from __future__ import annotations

import re
from pathlib import Path
from typing import TYPE_CHECKING

import mammos_units as u
import pandas as pd

import mammos_entity as me

if TYPE_CHECKING:
    import mammos_units
    import numpy.typing

    import mammos_entity


def entities_to_csv(
    filename: str | Path,
    **entities: mammos_entity.Entity | mammos_units.Quantity | numpy.typing.ArrayLike,
) -> None:
    """Write tabular data to csv file.

    The file structure is explained in the module-level documentation.

    Column names are keys of the dictionary. If an element is of type
    :py:class:`mammos_entity.Entity` ontology label, IRI and unit are added to the
    header, if an element is of type :py:class:`mammos_units.Quantity` its unit is added
    to the header, otherwise all headers are empty.

    """
    if not entities:
        raise RuntimeError("No data to write.")
    ontology_labels = []
    ontology_iris = []
    units = []
    data = {}
    if_scalar_list = []
    for name, element in entities.items():
        if isinstance(element, me.Entity):
            ontology_labels.append(element.ontology_label)
            ontology_iris.append(element.ontology.iri)
            units.append(str(element.unit))
            data[name] = element.value
            if_scalar_list.append(pd.api.types.is_scalar(element.value))
        elif isinstance(element, u.Quantity):
            ontology_labels.append("")
            ontology_iris.append("")
            units.append(str(element.unit))
            data[name] = element.value
            if_scalar_list.append(pd.api.types.is_scalar(element.value))
        else:
            ontology_labels.append("")
            ontology_iris.append("")
            units.append("")
            data[name] = element
            if_scalar_list.append(pd.api.types.is_scalar(element))

    dataframe = (
        pd.DataFrame(data, index=[0]) if all(if_scalar_list) else pd.DataFrame(data)
    )
    with open(filename, "w") as f:
        f.write("#mammos csv v1\n")
        f.write("#" + ",".join(ontology_labels) + "\n")
        f.write("#" + ",".join(ontology_iris) + "\n")
        f.write("#" + ",".join(units) + "\n")
        dataframe.to_csv(f, index=False)


class EntityCollection:
    """Container class storing entity-like objects."""

    def __init__(self, **kwargs):
        """Initialize EntityCollection, keywords become attributes of the class."""
        for key, val in kwargs.items():
            setattr(self, key, val)

    def __repr__(self):
        """Show container elements."""
        args = "\n".join(f"    {key}={val!r}," for key, val in self.__dict__.items())
        return f"{self.__class__.__name__}(\n{args}\n)"

    def to_dataframe(self, include_units: bool = True):
        """Convert values to dataframe."""

        def unit(key: str) -> str:
            """Get unit for element key.

            Returns:
                A string " (unit)" if the element has a unit, otherwise an empty string.
            """
            unit = getattr(getattr(self, key), "unit", None)
            if unit and str(unit):
                return f" ({unit!s})"
            else:
                return ""

        return pd.DataFrame(
            {
                f"{key}{unit(key) if include_units else ''}": getattr(val, "value", val)
                for key, val in self.__dict__.items()
            }
        )


def entities_from_csv(filename: str | Path) -> EntityCollection:
    """Read CSV file with ontology metadata.

    Reads a file as defined in the module description. The returned container provides
    access to the individual columns.
    """
    with open(filename) as f:
        file_version_information = f.readline()
        version = re.search(r"v\d+", file_version_information)
        if not version:
            raise RuntimeError("File does not have version information in line 1.")
        if version.group() != "v1":
            raise RuntimeError(
                f"Reading mammos csv {version.group()} is not supported."
            )
        ontology_labels = f.readline().removeprefix("#").removesuffix("\n").split(",")
        _ontology_iris = f.readline().removeprefix("#").removesuffix("\n").split(",")
        units = f.readline().removeprefix("#").removesuffix("\n").split(",")
        names = f.readline().removeprefix("#").removesuffix("\n").split(",")

        f.seek(0)
        data = pd.read_csv(f, comment="#", sep=",")
        scalar_data = len(data) == 1

    result = EntityCollection()

    for name, ontology_label, unit in zip(names, ontology_labels, units, strict=False):
        data_values = data[name].values if not scalar_data else data[name].values[0]
        if ontology_label:
            setattr(result, name, me.Entity(ontology_label, data_values, unit))
        elif unit:
            setattr(result, name, u.Quantity(data_values, unit))
        else:
            setattr(result, name, data_values)

    return result



"""Loads and provides access to MaMMoS ontology which is part of EMMO.

Loads and provides access to the MaMMoS magnetic materials ontology, including
everything from the EMMO ontology, via the `EMMOntoPy` library. The ontology is loaded
from ``.ttl`` (Turtle) files distributed with mammos-entity.
"""

from logging import getLogger
from pathlib import Path

import ontopy

logger = getLogger(__name__)


def load_offline_ontology() -> ontopy.ontology.Ontology:
    """Load EMMO and MaMMoS ontology from 'ontology' directory.

    The returned ontology object contains all definitions from both ontologies, EMMO is
    in the attribute `.imported_ontologies` and accessible in other methods when using
    ``imported=True``.

    """
    w = ontopy.World()
    ontology_dir = (Path(__file__).parent / "ontology").resolve()
    # load EMMO
    # using pathlib.Path(...).as_uri() causes ontopy to fail on Windows, therefore
    # we construct the file uri manually in the form required for ontopy
    emmo_ttl = f"file://{ontology_dir / 'emmo.ttl'!s}"
    logger.debug("loading emmo ttl from '%s'", emmo_ttl)
    w.get_ontology(emmo_ttl).load()
    # now load MaMMoS ontology, which builds upon EMMO; with EMMO already loaded
    # no internet access is required to resolve 'owl:imports <https://w3id.org/emmo'
    mammos_ttl = f"file://{ontology_dir / 'magnetic_material_mammos.ttl'!s}"
    logger.debug("loading mammos ttl from '%s'", mammos_ttl)
    mammos_ontology = w.get_ontology(mammos_ttl).load()
    return mammos_ontology


def load_online_ontology() -> ontopy.ontology.Ontology:
    """Fetch EMMO and MaMMoS ontology from the internet."""
    # We need to only load the MaMMoS ontology explicitly. Ontopy will automatically
    # fetch EMMO as dependency of the MaMMoS ontology.
    return ontopy.ontology.get_ontology(
        "https://raw.githubusercontent.com/MaMMoS-project/MagneticMaterialsOntology/refs/heads/main/magnetic_material_mammos.ttl"
    ).load()


mammos_ontology = load_offline_ontology()
